==> bdating_common-0.9.43-py3-none-any.whl: bdating_common/xero_helper.py <==
import base64
from xero_python.api_client import ApiClient, serialize
from xero_python.api_client.oauth2 import OAuth2Token
from xero_python.api_client.configuration import Configuration
from xero_python.accounting import AccountingApi, Contact, LineItem, Account, BankTransaction, BankTransactions, LineItem, Invoice, Invoices, Contacts
from functools import wraps
import jwt
from datetime import *
from xero_python.utils import getvalue
from xero_python.exceptions import AccountingBadRequestException
import dateutil
from dateutil.relativedelta import *
from .model import OrderCreatedRequest 
import decimal

client_id = "CC289C1597A44D20BD0C4C68AA4D52B0"
client_secret = "n6_p_yka9D3S5GZ_9Xg984-G4SbCw4TExr8TC-QKCLSUQ1K7"
tenent_id = "46b5b1a0-3ab4-4cdf-b684-b40228f17a0e"

message_bytes = (client_id + ":" + client_secret).encode("utf-8")
encoded_bytes = base64.b64encode(message_bytes)
encoded_message = encoded_bytes.decode("utf-8")

scope = "accounting.transactions accounting.transactions.read accounting.contacts accounting.contacts.read"

xero_token = {}

# configure xero-python sdk client
api_client = ApiClient(
    Configuration(
        debug=True,
        oauth2_token=OAuth2Token(
            client_id=client_id, client_secret=client_secret
        ),
    ),
    pool_threads=1,
)

def xero_token_required(function):
    @wraps(function)
    def decorator(*args, **kwargs):
        old_token = get_token()

        return function(*args, **kwargs)

    return decorator

@api_client.oauth2_token_saver
def set_token(token):
    claims = jwt.decode(token["access_token"], options={"verify_signature": False})
    token["expires_at"] = claims["exp"]
    global xero_token
    xero_token = token


@api_client.oauth2_token_getter
def get_token():
    global xero_token
    if xero_token == {} or xero_token['expires_at'] - datetime.now().timestamp() < 60:
        print ('expired or no token')
        temp_token = api_client.get_client_credentials_token()
        return temp_token
    return xero_token

@xero_token_required
def create_txns_on_order_confirmed(data: OrderCreatedRequest, referral_model: object):
    summarize_errors = 'True'
    api_instance = AccountingApi(api_client)

    # fetch contact id [provider_xero_id, ...referral_xero_ids]
    contact_ids = get_contactId_from_account_numbers(
        [data.p_uid] + data.referral_uid)

    platform_commission_gross = data.booking_value * referral_model.platform_commission_rate
    platform_commission = platform_commission_gross
    p_contribution = data.booking_value - platform_commission_gross

    #TODO: bank account
    bank_account = Account(
        account_id="98b9b91d-26f1-464d-9307-8a20ebb3d6e2")

    bank_transactions_array = []

    # iterate referrer
    for idx, ref in enumerate(data.referral_uid):
        # first id is provider xero id
        r_contact = Contact(contact_ids[idx + 1])
        reference_contribution = data.booking_value * referral_model.reference_rate[idx]
        platform_commission -= reference_contribution
        r_line_item = LineItem(
            description=f"bookingId={data.booking_id}, total={data.booking_value}, pId={data.p_uid}, rContribution={reference_contribution} in reference tier {idx+1}",
            quantity=1.0,
            unit_amount=reference_contribution,
            account_code="601")
        line_items = []
        line_items.append(r_line_item)

        bank_transaction = BankTransaction(
            type="RECEIVE",
            contact=r_contact,
            line_items=line_items,
            reference=f"referrer={data.booking_id}",
            bank_account=bank_account)

        bank_transactions_array.append(bank_transaction)

    # provider
    p_contact = Contact(contact_ids[0])
    p_line_item = LineItem(
        description=f"bookingId={data.booking_id}, total={data.booking_value}, pId={data.p_uid}, pContribution={p_contribution} ",
        quantity=1.0,
        unit_amount=p_contribution,
        account_code="601")
    line_items = []
    line_items.append(p_line_item)

    bank_transaction = BankTransaction(
        type="RECEIVE",
        contact=p_contact,
        line_items=line_items,
        reference=f"provider={data.booking_id}",
        bank_account=bank_account)

    bank_transactions_array.append(bank_transaction)

    # revenue
    bdating_contact = Contact("5f778246-9e76-418b-8444-25fd729ca378")
    revenue_line_item = LineItem(
        description=f"bookingId={data.booking_id}, total={data.booking_value}, pId={data.p_uid}, platform_commission={platform_commission} ",
        quantity=1.0,
        unit_amount=platform_commission,
        account_code="603")
    line_items = []
    line_items.append(revenue_line_item)

    bank_transaction = BankTransaction(
        type="RECEIVE",
        contact=bdating_contact,
        line_items=line_items,
        reference=f"revenue={data.booking_id}",
        bank_account=bank_account)

    bank_transactions_array.append(bank_transaction)

    bankTransactions = BankTransactions(
        bank_transactions=bank_transactions_array)

    try:
        api_response = api_instance.create_bank_transactions(
            xero_tenant_id=tenent_id, bank_transactions=bankTransactions, summarize_errors=summarize_errors, unitdp=4)
        return api_response
    except AccountingBadRequestException as e:
        print("Accounting: Exception when calling AccountingApi->createBankTransactions: %s\n" % e)
        raise e

@xero_token_required
def create_contact_on_registration(uid: str):
    api_instance = AccountingApi(api_client)
    summarize_errors = 'True'
    contact = Contact(
        name = uid,
        account_number = uid)
    contacts = Contacts( 
        contacts = [contact])
    try:
        api_response = api_instance.create_contacts(tenent_id, contacts, summarize_errors)
        return getvalue(api_response, 'contacts', '')
    except AccountingBadRequestException as e:
        print("Exception when calling AccountingApi->createContacts: %s\n" % e)
        raise e

@xero_token_required
def withdraw_request_received(
    account_id: str,
    amount: decimal.Decimal
):
    api_instance = AccountingApi(api_client)

    balance = check_balance(account_id=account_id)
    now = datetime.now()
    delta = relativedelta(weeks=1)
    due_date = now + delta
    if balance - amount >= 0:
        contactId = get_contactId_from_account_numbers(acc_nos=[account_id])[0]

        contact = Contact(
            contact_id=contactId)
        line_item = LineItem(
            description=f"With draw by user {contact.name} at amount {amount}, balance remaining {balance-amount}",
            quantity=1.0,
            unit_amount=amount,
            account_code="602")

        line_items = []
        line_items.append(line_item)

        bill = Invoice(
            type="ACCPAY",
            contact=contact,
            date=now,
            due_date=due_date,
            line_items=line_items,
            reference=f"Balance remaining={balance-amount}",
            status="AUTHORISED")

        invoices = Invoices(
            invoices=[bill])

        try:
            api_response = api_instance.create_invoices(
                xero_tenant_id=tenent_id, invoices=invoices, unitdp=4)
            return api_response
        except AccountingBadRequestException as e:
            print("Exception when calling AccountingApi->createInvoices: %s\n" % e)
            raise e
    else:
        raise PermissionError("No enough balance")

def check_balance(account_id: str):
    balance = decimal.Decimal(0)
    txns = get_txns(account_id=account_id)
    for txn in txns:
        balance += txn.total
    invoices = get_invoices(account_id=account_id)
    for invoice in invoices:
        balance -= invoice.total
    return balance

def get_invoices(account_id: str):
    api_instance = AccountingApi(api_client)
    where = f'(Status=="AUTHORISED" OR Status=="PAID") AND Contact.Name="{account_id}"'
    # created_by_my_app = 'False'
    # summary_only = 'True'

    try:
        api_response = api_instance.get_invoices(xero_tenant_id=tenent_id,
                                                 where=where,
                                                 page=0,
                                                 unitdp=4, )

        print(len(getvalue(api_response, "invoices", "")))
        return getvalue(api_response, "invoices", "")
    except AccountingBadRequestException as e:
        print("Exception when calling AccountingApi->getInvoices: %s\n" % e)

def get_txns(account_id: str):
    api_instance = AccountingApi(api_client)
    where = f'Type=="RECEIVE" AND Contact.Name=="{account_id}" AND Status=="AUTHORISED"'
    order = 'Date ASC'
    if_before = dateutil.parser.parse("2020-02-06T12:17:43.202-08:00")
    try:
        api_response = api_instance.get_bank_transactions(
            xero_tenant_id=tenent_id, if_modified_since=if_before, where=where, order=order, unitdp=4)
        txns = getvalue(api_response, 'bank_transactions', '')
        return txns
    except AccountingBadRequestException as e:
        print("Exception when calling AccountingApi->getBankTransactions: %s\n" % e)

def get_contactId_from_account_numbers(acc_nos: list):
    api_instance = AccountingApi(api_client)
    where = 'ContactStatus=="ACTIVE" AND '
    for id, acc in enumerate(acc_nos):
        if id != len(acc_nos)-1:
            where += 'AccountNumber=="' + acc + '" OR '
        else:
            where += 'AccountNumber=="' + acc + '"'

    order = 'Name ASC'
    include_archived = 'False'
    summary_only = 'True'

    try:
        api_response = api_instance.get_contacts(xero_tenant_id=tenent_id, where=where,
                                                 order=order, include_archived=include_archived, summary_only=summary_only)
        resp = getvalue(api_response, 'contacts', '')
        result = []
        for acc in acc_nos:
            for r in resp:
                if r.account_number == acc:
                    result.append(r.contact_id)
        return result
    except AccountingBadRequestException as e:
        print("Exception when calling AccountingApi->getContacts: %s\n" % e)

==> bdating_common-0.9.44-py3-none-any.whl: bdating_common/xero_helper.py <==
import base64
import logging
from xero_python.api_client import ApiClient, serialize
from xero_python.api_client.oauth2 import OAuth2Token
from xero_python.api_client.configuration import Configuration
from xero_python.accounting import AccountingApi, Contact, LineItem, Account, BankTransaction, BankTransactions, LineItem, Invoice, Invoices, Contacts
from functools import wraps
import jwt
from datetime import *
from xero_python.utils import getvalue
from xero_python.exceptions import AccountingBadRequestException
import dateutil
from dateutil.relativedelta import *
from .model import OrderCreatedRequest 
import decimal

log = logging.getLogger(__name__)

client_id = "CC289C1597A44D20BD0C4C68AA4D52B0"
client_secret = "n6_p_yka9D3S5GZ_9Xg984-G4SbCw4TExr8TC-QKCLSUQ1K7"
tenent_id = "46b5b1a0-3ab4-4cdf-b684-b40228f17a0e"

message_bytes = (client_id + ":" + client_secret).encode("utf-8")
encoded_bytes = base64.b64encode(message_bytes)
encoded_message = encoded_bytes.decode("utf-8")

scope = "accounting.transactions accounting.transactions.read accounting.contacts accounting.contacts.read"

xero_token = {}

# configure xero-python sdk client
api_client = ApiClient(
    Configuration(
        debug=True,
        oauth2_token=OAuth2Token(
            client_id=client_id, client_secret=client_secret
        ),
    ),
    pool_threads=1,
)

def xero_token_required(function):
    @wraps(function)
    def decorator(*args, **kwargs):
        old_token = get_token()

        return function(*args, **kwargs)

    return decorator

@api_client.oauth2_token_saver
def set_token(token):
    claims = jwt.decode(token["access_token"], options={"verify_signature": False})
    token["expires_at"] = claims["exp"]
    global xero_token
    xero_token = token


@api_client.oauth2_token_getter
def get_token():
    global xero_token
    if xero_token == {} or xero_token['expires_at'] - datetime.now().timestamp() < 60:
        print ('expired or no token')
        temp_token = api_client.get_client_credentials_token()
        return temp_token
    return xero_token
"""
referral_model
[
    {
        uid: 
        referral_config: {
            platform_commission: float
            reward_layers: list[float]
        }
    },
    ...
]
"""
@xero_token_required
def create_txns_on_order_confirmed(data: OrderCreatedRequest, provider_model: object, referral_models: list[object]):
    summarize_errors = 'True'
    api_instance = AccountingApi(api_client)

    referrer_uids = list(map(lambda m: m['uid'], referral_models))

    # fetch contact id [provider_xero_id, ...referral_xero_ids]
    contact_ids = get_contactId_from_account_numbers(
        [provider_model['uid']] + referrer_uids)
    
    xero_provider_id = contact_ids[0]
    xero_referal_ids = contact_ids[slice(1, -1)]

    platform_commission_gross = data.booking_value * provider_model['platform_commission_rate']
    platform_commission = platform_commission_gross
    p_contribution = data.booking_value - platform_commission_gross

    #TODO: bank account - bdating_wallet
    log.info("wallet === ", data['booking']['bdating_wallet'])
    bank_account = Account(
        account_id="98b9b91d-26f1-464d-9307-8a20ebb3d6e2")

    bank_transactions_array = []

    # iterate referrer
    for idx, ref in enumerate(referral_models):
        referral = referral_models[idx]
        # first id is provider xero id
        r_contact = Contact(xero_referal_ids[idx])
        reference_contribution = data.booking_value * referral['reward_layers'][idx]
        platform_commission -= reference_contribution
        r_line_item = LineItem(
            description=f"bookingId={data.booking_id}, total={data.booking_value}, referral={referrer_uids[idx]}, contribution={reference_contribution}, tier {idx+1}",
            quantity=1.0,
            unit_amount=reference_contribution,
            # account code
            account_code="601")
        line_items = []
        line_items.append(r_line_item)

        bank_transaction = BankTransaction(
            type="RECEIVE",
            contact=r_contact,
            line_items=line_items,
            reference=f"R{id+1}, booking_id={data.booking_id}",
            bank_account=bank_account)

        bank_transactions_array.append(bank_transaction)

    # provider
    p_contact = Contact(xero_provider_id)
    p_line_item = LineItem(
        description=f"bookingId={data.booking_id}, total={data.booking_value}, pId={data.p_uid}, pContribution={p_contribution} ",
        quantity=1.0,
        unit_amount=p_contribution,
        account_code="601")
    line_items = []
    line_items.append(p_line_item)

    bank_transaction = BankTransaction(
        type="RECEIVE",
        contact=p_contact,
        line_items=line_items,
        reference=f"P, booking_id={data.booking_id}",
        bank_account=bank_account)

    bank_transactions_array.append(bank_transaction)

    # revenue
    bdating_contact = Contact("5f778246-9e76-418b-8444-25fd729ca378")
    revenue_line_item = LineItem(
        description=f"bookingId={data.booking_id}, total={data.booking_value}, pId={data.p_uid}, platform_commission={platform_commission} ",
        quantity=1.0,
        unit_amount=platform_commission,
        account_code="603")
    line_items = []
    line_items.append(revenue_line_item)

    bank_transaction = BankTransaction(
        type="RECEIVE",
        contact=bdating_contact,
        line_items=line_items,
        reference=f"revenue={data.booking_id}",
        bank_account=bank_account)

    bank_transactions_array.append(bank_transaction)

    bankTransactions = BankTransactions(
        bank_transactions=bank_transactions_array)

    try:
        api_response = api_instance.create_bank_transactions(
            xero_tenant_id=tenent_id, bank_transactions=bankTransactions, summarize_errors=summarize_errors, unitdp=4)
        return api_response
    except AccountingBadRequestException as e:
        print("Accounting: Exception when calling AccountingApi->createBankTransactions: %s\n" % e)
        raise e

@xero_token_required
def create_contact_on_registration(uid: str):
    api_instance = AccountingApi(api_client)
    summarize_errors = 'True'
    contact = Contact(
        name = uid,
        account_number = uid)
    contacts = Contacts( 
        contacts = [contact])
    try:
        api_response = api_instance.create_contacts(tenent_id, contacts, summarize_errors)
        return getvalue(api_response, 'contacts', '')
    except AccountingBadRequestException as e:
        print("Exception when calling AccountingApi->createContacts: %s\n" % e)
        raise e

@xero_token_required
def withdraw_request_received(
    account_id: str,
    amount: decimal.Decimal
):
    api_instance = AccountingApi(api_client)

    balance = check_balance(account_id=account_id)
    now = datetime.now()
    delta = relativedelta(weeks=1)
    due_date = now + delta
    if balance - amount >= 0:
        contactId = get_contactId_from_account_numbers(acc_nos=[account_id])[0]

        contact = Contact(
            contact_id=contactId)
        line_item = LineItem(
            description=f"With draw by user {contact.name} at amount {amount}, balance remaining {balance-amount}",
            quantity=1.0,
            unit_amount=amount,
            account_code="602")

        line_items = []
        line_items.append(line_item)

        bill = Invoice(
            type="ACCPAY",
            contact=contact,
            date=now,
            due_date=due_date,
            line_items=line_items,
            reference=f"Balance remaining={balance-amount}",
            status="AUTHORISED")

        invoices = Invoices(
            invoices=[bill])

        try:
            api_response = api_instance.create_invoices(
                xero_tenant_id=tenent_id, invoices=invoices, unitdp=4)
            return api_response
        except AccountingBadRequestException as e:
            print("Exception when calling AccountingApi->createInvoices: %s\n" % e)
            raise e
    else:
        raise PermissionError("No enough balance")

def check_balance(account_id: str):
    balance = decimal.Decimal(0)
    txns = get_txns(account_id=account_id)
    for txn in txns:
        balance += txn.total
    invoices = get_invoices(account_id=account_id)
    for invoice in invoices:
        balance -= invoice.total
    return balance

def get_invoices(account_id: str):
    api_instance = AccountingApi(api_client)
    where = f'(Status=="AUTHORISED" OR Status=="PAID") AND Contact.Name="{account_id}"'
    # created_by_my_app = 'False'
    # summary_only = 'True'

    try:
        api_response = api_instance.get_invoices(xero_tenant_id=tenent_id,
                                                 where=where,
                                                 page=0,
                                                 unitdp=4, )

        print(len(getvalue(api_response, "invoices", "")))
        return getvalue(api_response, "invoices", "")
    except AccountingBadRequestException as e:
        print("Exception when calling AccountingApi->getInvoices: %s\n" % e)

def get_txns(account_id: str):
    api_instance = AccountingApi(api_client)
    where = f'Type=="RECEIVE" AND Contact.Name=="{account_id}" AND Status=="AUTHORISED"'
    order = 'Date ASC'
    if_before = dateutil.parser.parse("2020-02-06T12:17:43.202-08:00")
    try:
        api_response = api_instance.get_bank_transactions(
            xero_tenant_id=tenent_id, if_modified_since=if_before, where=where, order=order, unitdp=4)
        txns = getvalue(api_response, 'bank_transactions', '')
        return txns
    except AccountingBadRequestException as e:
        print("Exception when calling AccountingApi->getBankTransactions: %s\n" % e)

def get_contactId_from_account_numbers(acc_nos: list):
    api_instance = AccountingApi(api_client)
    where = 'ContactStatus=="ACTIVE" AND '
    for id, acc in enumerate(acc_nos):
        if id != len(acc_nos)-1:
            where += 'AccountNumber=="' + acc + '" OR '
        else:
            where += 'AccountNumber=="' + acc + '"'

    order = 'Name ASC'
    include_archived = 'False'
    summary_only = 'True'

    try:
        api_response = api_instance.get_contacts(xero_tenant_id=tenent_id, where=where,
                                                 order=order, include_archived=include_archived, summary_only=summary_only)
        resp = getvalue(api_response, 'contacts', '')
        result = []
        for acc in acc_nos:
            for r in resp:
                if r.account_number == acc:
                    result.append(r.contact_id)
        return result
    except AccountingBadRequestException as e:
        print("Exception when calling AccountingApi->getContacts: %s\n" % e)
